package eu.clarin.cqi.ws.impl;

import de.tuebingen.uni.sfs.wlf1.tc.api.*;
import de.tuebingen.uni.sfs.wlf1.tc.xb.TextCorpusStored;
import de.tuebingen.uni.sfs.wlf1.xb.WLData;
import eu.clarin.cqi.client.CqiClient;
import eu.clarin.cqi.client.CqiClientException;
import eu.clarin.cqi.client.CqiResult;
import java.util.*;

/**
 *
 * @author akislev
 */
public class QueryEngine {

    private static final String CONTEXT_STRUCTURAL_ATTRIBUTE = "s";
    private static final String LANG = "de";
    private static final String TAG_SET = "stts";
    private final CqiClient client;

    public QueryEngine(String serverHost, int serverPort, String username, String password) throws QueryEngineException {
        try {
            client = new CqiClient(serverHost, serverPort);
        } catch (CqiClientException ex) {
            throw new QueryEngineException("can't initialize a cqi client", ex);
        }
        try {
            client.connect(username, password);
        } catch (CqiClientException ex) {
            throw new QueryEngineException("can't connect to the cqi server", ex);
        }
    }

    public void close() throws QueryEngineException {
        try {
            client.disconnect();
        } catch (CqiClientException ex) {
            throw new QueryEngineException("can't disconnect from the cqi server", ex);
        }
    }

    public WLData query(String corpus, String query, int startRecord, int maximumMatches) throws QueryEngineException, CqiClientException {
        String[] mdStructAttr = getStructuralAttributeWithValues(corpus);
        CqiResult result = client.query(corpus, query, CONTEXT_STRUCTURAL_ATTRIBUTE, mdStructAttr);
        if (result.size() == 0) {
            throw new QueryEngineException("No results found.");
        }
        if (maximumMatches < 1) {
            result.clear();
            throw new QueryEngineException("too few matches requested " + maximumMatches);
        }
        if (result.size() != 0) {
            if (startRecord < 1 || startRecord > result.size()) {
                throw new QueryEngineException("First record position out of range.");
            }
        }
        String wordAttr = "word";
        String lemmaAttr = "lemma";
        String posAttr = "pos";
        for (String attr : client.corpusPositionalAttributes(corpus)) {
            if (attr.contains(wordAttr)) {
                wordAttr = attr;
            }
            if (attr.contains(lemmaAttr)) {
                lemmaAttr = attr;
            }
            if (attr.contains(posAttr)) {
                posAttr = attr;
            }
        }
        TextCorpusStored textCorpus = new TextCorpusStored(LANG);
        TextLayer textLayer = textCorpus.createTextLayer();
        TokensLayer tokensLayer = textCorpus.createTokensLayer();
        SentencesLayer sentenceLayer = textCorpus.createSentencesLayer();
        LemmasLayer lemmasLayer = textCorpus.createLemmasLayer();
        PosTagsLayer posTagsLayer = textCorpus.createPosTagsLayer(TAG_SET);
        MatchesLayer matchesLayer = textCorpus.createMatchesLayer("cqp", query);
        StringBuilder sb = new StringBuilder();
        boolean isFirst = true;
        MatchedCorpus matchedCorpus = matchesLayer.addCorpus(corpus, null);
        if (result.absolute(startRecord - 1)) {
            do {
                int matchStart = result.getMatchStart();
                int matchEnd = result.getMatchEnd();
                int contextStart = result.getContextStart();
                int contextEnd = result.getContextEnd();
                int target = result.getTarget();
                int leftContextSize = matchStart - contextStart;
                int targetOffset;
                if (target != 0) {
                    targetOffset = target - contextStart;
                } else {
                    targetOffset = -1;
                }
                int relMatchEnd = matchEnd - contextStart;
                String[] words = result.getValues(wordAttr, contextStart, contextEnd);
                String[] lemmas = result.getValues(lemmaAttr, contextStart, contextEnd);
                String[] tags = result.getValues(posAttr, contextStart, contextEnd);
                List<Token> sentence = new ArrayList<Token>(contextEnd - contextStart + 1);
                List<Token> matchedTokens = new ArrayList<Token>(matchEnd - matchStart + 1);
                Token targetToken = null;
                for (int j = 0; j < words.length; j++) {
                    String word = words[j];
                    Token token = tokensLayer.addToken(word);
                    sentence.add(token);
                    if (isFirst) {
                        isFirst = false;
                    } else {
                        sb.append(' ');
                    }
                    sb.append(word);
                    lemmasLayer.addLemma(lemmas[j], token);
                    posTagsLayer.addTag(tags[j], token);
                    if (j >= leftContextSize && j <= relMatchEnd) {
                        matchedTokens.add(token);
                    }
                    if (j == targetOffset) {
                        targetToken = token;
                    }
                }
                sentenceLayer.addSentence(sentence);
                Map<String, String> targets = new HashMap<String, String>(1);
                if (targetToken != null) {
                    targets.put("0", targetToken.getID());
                }
                Map<String, String> metadata = new HashMap<String, String>(mdStructAttr.length);
                for (int i = 0; i < mdStructAttr.length; i++) {
                    String value = result.getStructuralAttributeValue(i);
                    if (!value.isEmpty()) {
                        metadata.put(mdStructAttr[i], value);
                    }
                }
                matchesLayer.addItem(matchedCorpus, matchedTokens, createIntList(matchStart, matchEnd), targets, metadata);
                if (--maximumMatches == 0) {
                    break;
                }
            } while (result.next());
            result.clear();
            textLayer.addText(sb.toString());
            return new WLData(textCorpus);
        }
        return null;
    }

    private static List<String> createIntList(int from, int to) {
        String[] ints = new String[to - from + 1];
        for (int i = 0; i < ints.length; i++) {
            ints[i] = Integer.toString(from + i);
        }
        return Arrays.asList(ints);
    }

    private String[] getStructuralAttributeWithValues(String corpus) throws CqiClientException {
        List<String> attrs = new ArrayList<String>();
        String[] structuralAttributes = client.corpusStructuralAttributes(corpus);
        for (String attr : structuralAttributes) {
            String attrName = String.format("%s.%s", corpus, attr);
            if (client.corpusStructuralAttributeHasValues(attrName)) {
                attrs.add(attr);
            }
        }
        return attrs.toArray(new String[attrs.size()]);
    }
}
